/*
 * Database record class for DMAP sharing
 *
 * Copyright (C) 2008 W. Michael Petullo <mike@flyn.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "dmapd-dmap-container-record.h"

enum {
	PROP_0,
	PROP_NAME
};

/* Media ID's start at max and go down.
 * Container ID's start at 1 and go up.
 * 1 is reserved for the primary container.
 */
static gint nextid = 2;

struct DmapdDMAPContainerRecordPrivate {
	gint id;
	const char *name;
	DMAPDb *entries;
};

static void
dmapd_dmap_container_record_set_property (GObject *object,
					  guint prop_id,
					  const GValue *value,
					  GParamSpec *pspec)
{
	DmapdDMAPContainerRecord *record = DMAPD_DMAP_CONTAINER_RECORD (object);
	GObject *obj;

	switch (prop_id) {
		case PROP_NAME:
			/* FIXME: should I free name first? */
			record->priv->name = g_value_dup_string (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

static void
dmapd_dmap_container_record_get_property (GObject *object,
					  guint prop_id,
					  GValue *value,
					  GParamSpec *pspec)
{
	DmapdDMAPContainerRecord *record = DMAPD_DMAP_CONTAINER_RECORD (object);
	GObject *obj;

	switch (prop_id) {
		case PROP_NAME:
			g_value_set_string (value, record->priv->name);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
			break;
	}
}

gint
dmapd_dmap_container_record_get_id (DMAPContainerRecord *record)
{
	return DMAPD_DMAP_CONTAINER_RECORD (record)->priv->id;
}

void
dmapd_dmap_container_record_add_entry (DMAPContainerRecord *container_record,
				       DMAPRecord *record, gint id)
{
	dmapd_dmap_db_add_with_id (DMAPD_DMAP_CONTAINER_RECORD 
				(container_record)->priv->entries, 
				record, id);
}

guint64
dmapd_dmap_container_record_get_entry_count (DMAPContainerRecord *record)
{
	return dmap_db_count (DMAPD_DMAP_CONTAINER_RECORD (record)->priv->entries);
}

const DMAPDb *
dmapd_dmap_container_record_get_entries (DMAPContainerRecord *record)
{
	return DMAPD_DMAP_CONTAINER_RECORD (record)->priv->entries;
}

static void
dmapd_dmap_container_record_init (DmapdDMAPContainerRecord *record)
{
	record->priv = DMAPD_DMAP_CONTAINER_RECORD_GET_PRIVATE (record);
	record->priv->id = nextid++;
	record->priv->entries = DMAP_DB (dmapd_dmap_db_new ());
}

static void
dmapd_dmap_container_record_class_init (DmapdDMAPContainerRecordClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

        g_type_class_add_private (klass, sizeof (DmapdDMAPContainerRecordPrivate));

	gobject_class->set_property = dmapd_dmap_container_record_set_property;
	gobject_class->get_property = dmapd_dmap_container_record_get_property;

	g_object_class_override_property (gobject_class, PROP_NAME, "name");
}

static void
dmapd_dmap_container_record_interface_init (gpointer iface, gpointer data)
{
	DMAPContainerRecordInterface *dmap_container_record = iface;

	g_assert (G_TYPE_FROM_INTERFACE (dmap_container_record) == TYPE_DMAP_CONTAINER_RECORD);

	dmap_container_record->get_id = dmapd_dmap_container_record_get_id;
	dmap_container_record->add_entry = dmapd_dmap_container_record_add_entry;
	dmap_container_record->get_entry_count = dmapd_dmap_container_record_get_entry_count;
	dmap_container_record->get_entries = dmapd_dmap_container_record_get_entries;
}

G_DEFINE_TYPE_WITH_CODE (DmapdDMAPContainerRecord, dmapd_dmap_container_record, G_TYPE_OBJECT, 
			G_IMPLEMENT_INTERFACE (TYPE_DMAP_CONTAINER_RECORD,
					       dmapd_dmap_container_record_interface_init))
