/*
 *  Database class for DMAP sharing
 *
 *  Copyright (C) 2008 W. Michael Petullo <mike@flyn.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <glib/ghash.h>
#include <libdmapsharing/dmap.h>

#include "dmapd-dmap-db.h"

struct DmapdDMAPDbPrivate {
	GHashTable *db;
	guint nextid;
};

DMAPRecord *dmapd_dmap_db_lookup_by_id	(DMAPDb *db, guint id)
{
	DMAPRecord *record;
	record = g_hash_table_lookup (DMAPD_DMAP_DB (db)->priv->db, GUINT_TO_POINTER (id));
	g_object_ref (record);
	return record;
}

void dmapd_dmap_db_foreach	(const DMAPDb *db,
				 GHFunc func,
				 gpointer data)
{
	g_hash_table_foreach (DMAPD_DMAP_DB (db)->priv->db, func, data);
}

gint64 dmapd_dmap_db_count (const DMAPDb *db)
{
	return g_hash_table_size (DMAPD_DMAP_DB (db)->priv->db);
}

gint dmapd_dmap_db_add (DMAPDb *db, DMAPRecord *record)
{
	guint id;
	id = DMAPD_DMAP_DB (db)->priv->nextid--;
	g_object_ref (record);
	g_hash_table_insert (DMAPD_DMAP_DB (db)->priv->db, GUINT_TO_POINTER (id), record);
	return id;
}

gint dmapd_dmap_db_add_with_id (DMAPDb *db, DMAPRecord *record, guint id)
{
	g_hash_table_insert (DMAPD_DMAP_DB (db)->priv->db, GUINT_TO_POINTER (id), record);
	return id;
}

static void dmapd_dmap_db_init (DmapdDMAPDb *db)
{
	db->priv = DMAPD_DMAP_DB_GET_PRIVATE (db);
	db->priv->db = g_hash_table_new (g_direct_hash, g_direct_equal);

	/* Media ID's start at max and go down.
	 * Container ID's start at 1 and go up.
	 */
	db->priv->nextid = G_MAXINT;
}

static void dmapd_dmap_db_class_init (DmapdDMAPDbClass *klass)
{
	g_type_class_add_private (klass, sizeof (DmapdDMAPDbPrivate));
}

static void dmapd_dmap_db_interface_init (gpointer iface, gpointer data)
{
	DMAPDbInterface *dmap_db = iface;

	g_assert (G_TYPE_FROM_INTERFACE (dmap_db) == TYPE_DMAP_DB);

	dmap_db->add = dmapd_dmap_db_add;
	dmap_db->lookup_by_id = dmapd_dmap_db_lookup_by_id;
	dmap_db->foreach = dmapd_dmap_db_foreach;
	dmap_db->count = dmapd_dmap_db_count;
}

G_DEFINE_TYPE_WITH_CODE (DmapdDMAPDb, dmapd_dmap_db, G_TYPE_OBJECT, 
			 G_IMPLEMENT_INTERFACE (TYPE_DMAP_DB, dmapd_dmap_db_interface_init))

DmapdDMAPDb *dmapd_dmap_db_new (void)
{
	DmapdDMAPDb *db;

	db = DMAPD_DMAP_DB (g_object_new (TYPE_DMAPD_DMAP_DB, NULL));

	return db;
}
